from typing import TYPE_CHECKING
from urllib.request import getproxies

from fuo_ytmusic.language import resolve_language

if TYPE_CHECKING:
    from feeluown.app import App

__alias__ = "ytmusic"
__version__ = "0.4.18"
__desc__ = "YouTube Music plugin"


ui_mgr = None


def init_config(config):
    # For example: http://127.0.0.1:7890. This will be used in API and media accessing.
    config.deffield("HTTP_PROXY", type_=str, default="", desc="YouTube Music HTTP proxy")
    config.deffield("HTTP_TIMEOUT", type_=int, default=2, desc="HTTP requests timeout")
    config.deffield(
        "LANGUAGE",
        type_=str,
        default="auto",
        desc="YouTube Music language (auto to follow system)",
    )


def enable(app: "App"):
    global ui_mgr

    from fuo_ytmusic.provider import provider

    # Use system http proxy by default.
    sys_proxies = getproxies()
    sys_http_proxy = sys_proxies.get("http")
    config_http_proxy = app.config.ytmusic.HTTP_PROXY
    if not config_http_proxy and sys_http_proxy:
        config_http_proxy = sys_http_proxy

    provider.setup_http_proxy(config_http_proxy)
    provider.setup_http_timeout(app.config.ytmusic.HTTP_TIMEOUT)
    provider.setup_language(resolve_language(app, app.config.ytmusic.LANGUAGE))
    app.library.register(provider)
    if app.mode & app.GuiMode:
        from .provider_ui import ProviderUI

        provider_ui = ProviderUI(app)
        app.pvd_ui_mgr.register(provider_ui)


def disable(app: "App"):
    global ui_mgr

    provider = app.library.get("ytmusic")
    if provider is not None:
        app.library.deregister(provider)
    ui_mgr = None



import importlib
import os

# 新版 FeelUOwn 会在启动的时候，设置这个环境变量
qt_moduel_name = os.getenv("FEELUOWN_QT_API", "PyQt5")

if qt_moduel_name != "PyQt5":
    qt_moduel_name = "PyQt6"

QtCore = importlib.import_module(qt_moduel_name + ".QtCore", package=qt_moduel_name)
QtGui = importlib.import_module(qt_moduel_name + ".QtGui", package=qt_moduel_name)
QtWidgets = importlib.import_module(
    qt_moduel_name + ".QtWidgets", package=qt_moduel_name
)

QLineEdit = QtWidgets.QLineEdit
QLabel = QtWidgets.QLabel
QPushButton = QtWidgets.QPushButton
QDialog = QtWidgets.QDialog
QVBoxLayout = QtWidgets.QVBoxLayout
QFormLayout = QtWidgets.QFormLayout
QFileDialog = QtWidgets.QFileDialog
QMessageBox = QtWidgets.QMessageBox
QInputDialog = QtWidgets.QInputDialog

pyqtSignal = QtCore.pyqtSignal

if qt_moduel_name == "PyQt5":
    RichText = QtCore.Qt.RichText
    Password = QLineEdit.Password
    QAction = QtWidgets.QAction
    TextSelectableByMouse = QtCore.Qt.TextSelectableByMouse
    ActionsContextMenu = QtCore.Qt.ActionsContextMenu
    Dialog = QtCore.Qt.Dialog
else:
    RichText = QtCore.Qt.TextFormat.RichText
    Password = QLineEdit.EchoMode.Password
    QAction = QtGui.QAction
    TextSelectableByMouse = QtCore.Qt.TextInteractionFlag.TextSelectableByMouse
    ActionsContextMenu = QtCore.Qt.ContextMenuPolicy.ActionsContextMenu
    Dialog = QtCore.Qt.WindowType.Dialog



import logging
import ntpath
import os
import sys
import threading
from datetime import timedelta
from enum import Enum
from functools import partial
from http.cookies import SimpleCookie
from typing import List, Optional, Union

import requests
from cachetools.func import ttl_cache
from feeluown.library import SearchType
from requests import Response
from ytmusicapi import YTMusic as YTMusicBase
from ytmusicapi.ytmusic import OAuthCredentials

from fuo_ytmusic.headerfile import (
    update_headerfile_cookie,
)
from fuo_ytmusic.helpers import Singleton
from fuo_ytmusic.lyrics import (
    build_watch_playlist_body,
    extract_lyrics_browse_id,
    timestamped_lyrics_to_lrc,
)
from fuo_ytmusic.models import (
    AlbumInfo,
    ArtistInfo,
    Categories,
    PlaylistAddItemResponse,
    PlaylistInfo,
    PlaylistNestedResult,
    SongInfo,
    UserInfo,
    YtmusicDispatcher,
    YtmusicHistorySong,
    YtmusicLibraryArtist,
    YtmusicLibrarySong,
    YtmusicSearchAlbum,
    YtmusicSearchArtist,
    YtmusicSearchBase,
    YtmusicSearchPlaylist,
    YtmusicSearchSong,
    YtmusicSearchVideo,
)
from fuo_ytmusic.profile import YtmusicProfileManager

CACHE_TTL = timedelta(minutes=10).seconds
CACHE_SIZE = 1
GLOBAL_LIMIT = 20

logger = logging.getLogger(__name__)


class YtmusicType(Enum):
    so = "songs"
    vi = "videos"
    ar = "artists"
    al = "albums"
    pl = "playlists"

    # noinspection PyTypeChecker
    @classmethod
    def parse(cls, type_: SearchType) -> "YtmusicType":
        return cls._value2member_map_.get(type_.value + "s")


class YtmusicPrivacyStatus(Enum):
    PUBLIC = "PUBLIC"
    PRIVATE = "PRIVATE"
    UNLISTED = "UNLISTED"

    # noinspection PyTypeChecker
    @classmethod
    def parse(cls, type_: str) -> "YtmusicPrivacyStatus":
        return cls._value2member_map_.get(type_)


class YtmusicScope(Enum):
    li = "library"
    up = "uploads"


class YTMusic(YTMusicBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.headerfile_path = None

    def set_headerfile_path(self, headerfile_path):
        self.headerfile_path = headerfile_path

    def check_auth(self):
        self._check_auth()

    def set_on_behalf_of_user(self, gaia_id):
        context = self.context.setdefault("context", {})
        user_ctx = context.setdefault("user", {})
        if gaia_id:
            user_ctx["onBehalfOfUser"] = gaia_id
        else:
            user_ctx.pop("onBehalfOfUser", None)

    def send_api_request(self, endpoint: str, body: dict):
        return self._send_request(endpoint, body)

    def request_with_auth(
        self, method: str, url: str, origin: str = None, json_body=None
    ):
        headers = dict(self.headers)
        if origin:
            headers["origin"] = origin
        return self._session.request(
            method,
            url,
            json=json_body,
            headers=headers,
            proxies=self._session.proxies,
            cookies=self.cookies,
        )

    def update_auth_cookie_from_response(self, response):
        cookie_header = self._auth_headers.get("cookie", "")
        if not response.cookies:
            return
        jar = SimpleCookie()
        if cookie_header:
            jar.load(cookie_header)
        for cookie in response.cookies:
            jar[cookie.name] = cookie.value
        new_cookie = "; ".join([f"{m.key}={m.value}" for m in jar.values()])
        if new_cookie:
            self._auth_headers["cookie"] = new_cookie
            self._persist_cookie_to_headerfile(new_cookie)

    def _persist_cookie_to_headerfile(self, cookie_value: str):
        update_headerfile_cookie(cookie_value, self.headerfile_path)

    class IterableToFileAdapter(object):
        def __init__(self, iterable):
            self.iterator = iter(iterable)
            self.length = len(iterable)

        def read(self, size=-1):  # TBD: add buffer for `len(data) > size` case
            return next(self.iterator, b"")

        def __len__(self):
            return self.length

    class ChunckedUpload(object):
        def __init__(self, filename, chunksize=1 << 6):
            self.filename = filename
            self.chunksize = chunksize
            self.totalsize = os.path.getsize(filename)
            self.readsofar = 0

        def __iter__(self):
            with open(self.filename, "rb") as file:
                while True:
                    data = file.read(self.chunksize)
                    if not data:
                        sys.stdout.write("\n")
                        break
                    self.readsofar += len(data)
                    percent = self.readsofar * 1e2 / self.totalsize
                    sys.stdout.write(
                        "\rUploading: {percent:3.0f}%".format(percent=percent)
                    )
                    yield data

        def __len__(self):
            return self.totalsize

    def upload_song_progress(self, filepath: str) -> Union[str, requests.Response]:
        """
        Uploads a song to YouTube Music

        :param filepath: Path to the music file (mp3, m4a, wma, flac or ogg)
        :return: Status String or full response
        """
        self._check_auth()
        if not os.path.isfile(filepath):
            raise Exception("The provided file does not exist.")

        supported_filetypes = ["mp3", "m4a", "wma", "flac", "ogg"]
        if os.path.splitext(filepath)[1][1:] not in supported_filetypes:
            raise Exception(
                "The provided file type is not supported by YouTube Music."
                f" Supported file types are {', '.join(supported_filetypes)}"
            )

        headers = self.headers.copy()
        upload_url = "https://upload.youtube.com/upload/usermusic/http?authuser=0"
        filesize = os.path.getsize(filepath)
        body = ("filename=" + ntpath.basename(filepath)).encode("utf-8")
        headers.pop("content-encoding", None)
        headers["content-type"] = "application/x-www-form-urlencoded;charset=utf-8"
        headers["X-Goog-Upload-Command"] = "start"
        headers["X-Goog-Upload-Header-Content-Length"] = str(filesize)
        headers["X-Goog-Upload-Protocol"] = "resumable"
        response = requests.post(
            upload_url, data=body, headers=headers, proxies=self.proxies
        )
        headers["X-Goog-Upload-Command"] = "upload, finalize"
        headers["X-Goog-Upload-Offset"] = "0"
        upload_url = response.headers["X-Goog-Upload-URL"]
        response = requests.post(
            upload_url,
            data=YTMusic.IterableToFileAdapter(YTMusic.ChunckedUpload(filepath)),
            headers=headers,
            proxies=self.proxies,
        )
        if response.status_code == 200:
            return "STATUS_SUCCEEDED"
        else:
            return response


class YtmusicService(metaclass=Singleton):
    def __init__(self):
        self._session = requests.Session()
        self._api: Optional[YTMusic] = None
        self._anonymous_lyrics_api: Optional[YTMusic] = None
        self._timeout: Optional[int] = None
        self._session.hooks["response"].append(self._do_logging)

        self._signature_timestamp = 0
        self._api_lock = threading.Lock()
        self._anonymous_lyrics_api_lock = threading.Lock()
        self._profile_manager = YtmusicProfileManager(self)
        self._language: Optional[str] = None

    @staticmethod
    def _do_logging(r: Response, *_, **__):
        logger.debug(
            f"[ytmusic] Requesting: [{r.request.method.upper()}] {r.url}; "
            f"Response: [{r.status_code}] {len(r.content)} bytes."
        )

    @property
    def api(self) -> YTMusic:
        if self._api is None:
            with self._api_lock:
                if self._api is None:
                    self.reinitialize_by_headerfile()
        return self._api

    def _log_thread(self):
        return f"Thread({threading.get_ident()})"

    def get_signature_timestamp(self):
        return 0

    def reinitialize_by_headerfile(self, headerfile=None):
        self._api = self._create_api(headerfile, self._session)

    def _create_api(self, headerfile, session: requests.Session) -> YTMusic:
        language = self._language or "zh_CN"
        options = dict(
            requests_session=session,
            language=language,
            oauth_credentials=OAuthCredentials(
                # In the new version of ytmusicapi, client_id and client_secret
                # are required args.
                #   https://github.com/sigma67/ytmusicapi/pull/688
                # Hardcode the client_id and client_secret to workaround.
                client_id=(
                    "861556708454-d6dlm3lh05idd8npek18k6be8ba3oc68"
                    ".apps.googleusercontent.com"
                ),
                client_secret="SboVhoG9s0rNafixCSGGKXAT",
                session=session,
            ),
        )
        # Due to https://github.com/sigma67/ytmusicapi/issues/676,
        # YTMusic does not work in specific cases when auth file is provided.
        # So initialize without auth file when 400 is returned.
        if headerfile is not None and headerfile.exists():
            logger.info("Initializing ytmusic api with headerfile.")
            api = YTMusic(str(headerfile), **options)
            api.set_headerfile_path(headerfile)
        else:
            logger.info("Initializing ytmusic api with no headerfile.")
            api = YTMusic(**options)
        return api

    def _get_anonymous_lyrics_api(self) -> YTMusic:
        if self._anonymous_lyrics_api is None:
            with self._anonymous_lyrics_api_lock:
                if self._anonymous_lyrics_api is None:
                    # Headerfile-authenticated clients can get HTTP 400 when
                    # requesting timestamped lyrics, while anonymous requests
                    # for the same lyrics browse id succeed.
                    session = requests.Session()
                    session.hooks["response"].append(self._do_logging)
                    session.proxies = dict(self._session.proxies)
                    if self._timeout is not None:
                        session.request = partial(
                            session.request,
                            timeout=self._timeout,
                        )
                    self._anonymous_lyrics_api = self._create_api(None, session)
        return self._anonymous_lyrics_api

    def _clear_anonymous_lyrics_api(self):
        with self._anonymous_lyrics_api_lock:
            self._anonymous_lyrics_api = None

    def setup_language(self, language: str):
        self._language = language
        self._clear_anonymous_lyrics_api()

    def setup_http_proxy(self, http_proxy):
        self._session.proxies = {
            "http": http_proxy,
            "https": http_proxy,
        }
        self._clear_anonymous_lyrics_api()

    def setup_timeout(self, timeout):
        self._timeout = timeout
        if isinstance(self._session.request, partial):
            request = self._session.request.func
        else:
            request = self._session.request
        self._session.request = partial(request, timeout=timeout)
        self._clear_anonymous_lyrics_api()

    def search(
        self,
        keywords: str,
        t: Optional[YtmusicType],
        scope: YtmusicScope = None,
        page_size: int = GLOBAL_LIMIT,
    ) -> List[
        Union[
            YtmusicSearchSong,
            YtmusicSearchAlbum,
            YtmusicSearchArtist,
            YtmusicSearchVideo,
            YtmusicSearchPlaylist,
            YtmusicSearchBase,
        ]
    ]:
        response = self.api.search(
            keywords,
            None if t is None else t.value,
            None if scope is None else scope.value,
            page_size,
        )
        return [YtmusicDispatcher.search_result_dispatcher(**data) for data in response]

    def _daily_home_cache_key(self) -> str:
        context = getattr(self.api, "context", {})
        if not isinstance(context, dict):
            return ""
        user_ctx = context.get("context", {}).get("user", {})
        if not isinstance(user_ctx, dict):
            return ""
        return str(user_ctx.get("onBehalfOfUser") or "")

    @ttl_cache(maxsize=8, ttl=30)
    def _home_sections_cached(self, limit: int, account_key: str):
        return self.api.get_home(limit)

    def home_sections(self, limit: int = 12):
        return self._home_sections_cached(limit, self._daily_home_cache_key())

    def get_current_account_info(self) -> dict:
        return self._profile_manager.get_current_account_info()

    def list_profiles(self) -> List[dict]:
        return self._profile_manager.list_profiles()

    def switch_profile(self, account_name: str = None, gaia_id: str = None) -> dict:
        return self._profile_manager.switch_profile(
            account_name=account_name, gaia_id=gaia_id
        )

    @ttl_cache(maxsize=CACHE_SIZE, ttl=CACHE_TTL)
    def artist_info(self, channel_id: str) -> ArtistInfo:
        data = self.api.get_artist(channel_id)
        return ArtistInfo(**data)

    @ttl_cache(maxsize=CACHE_SIZE, ttl=CACHE_TTL)
    def artist_albums(self, channel_id: str, params: str) -> List[YtmusicSearchAlbum]:
        response = self.api.get_artist_albums(channel_id, params)
        return [YtmusicSearchAlbum(**data) for data in response]

    @ttl_cache(maxsize=CACHE_SIZE, ttl=CACHE_TTL)
    def user_info(self, channel_id: str) -> UserInfo:
        return UserInfo(**self.api.get_user(channel_id))

    def user_playlists(self, channel_id: str, params: str):
        return self.api.get_user_playlists(channel_id, params)

    @ttl_cache(maxsize=CACHE_SIZE, ttl=CACHE_TTL)
    def album_info(self, browse_id: str) -> AlbumInfo:
        data = self.api.get_album(browse_id)
        return AlbumInfo(**data)

    def song_info(self, video_id: str) -> SongInfo:
        return SongInfo(**self.api.get_song(video_id, self.get_signature_timestamp()))

    def _song_lyrics_browse_id(self, video_id: str, api=None) -> Optional[str]:
        if api is None:
            api = self.api
        response = api.send_api_request("next", build_watch_playlist_body(video_id))
        return extract_lyrics_browse_id(response)

    def song_lyrics(self, video_id: str) -> Optional[str]:
        lyrics_api = self._get_anonymous_lyrics_api()
        lyrics_browse_id = self._song_lyrics_browse_id(video_id, api=lyrics_api)
        if lyrics_browse_id is None:
            return None
        lyrics_payload = lyrics_api.get_lyrics(lyrics_browse_id, timestamps=True)
        return timestamped_lyrics_to_lrc(lyrics_payload)

    @ttl_cache(maxsize=CACHE_SIZE, ttl=CACHE_TTL)
    def categories(self) -> List[Categories]:
        return [
            Categories(key=k, value=v)
            for k, v in self.api.get_mood_categories().items()
        ]

    @ttl_cache(maxsize=CACHE_SIZE, ttl=CACHE_TTL)
    def category_playlists(self, params: str) -> List[PlaylistNestedResult]:
        response = self.api.get_mood_playlists(params)
        return [PlaylistNestedResult(**data) for data in response]

    @ttl_cache(maxsize=CACHE_SIZE, ttl=CACHE_TTL)
    def get_charts(self, country: str = "ZZ") -> dict:
        response = self.api.get_charts(country)
        return response if isinstance(response, dict) else {}

    def library_playlists(
        self, limit: int = GLOBAL_LIMIT
    ) -> List[PlaylistNestedResult]:
        response = self.api.get_library_playlists(limit)
        return [PlaylistNestedResult(**data) for data in response]

    def library_songs(self, limit: int = GLOBAL_LIMIT) -> List[YtmusicLibrarySong]:
        response = self.api.get_library_songs(limit)
        return [YtmusicLibrarySong(**data) for data in response]

    def library_albums(self, limit: int = GLOBAL_LIMIT) -> List[YtmusicSearchAlbum]:
        response = self.api.get_library_albums(limit)
        return [YtmusicSearchAlbum(**data) for data in response]

    def library_artists(self, limit: int = GLOBAL_LIMIT) -> List[YtmusicLibraryArtist]:
        response = self.api.get_library_artists(limit)
        return [YtmusicLibraryArtist(**data) for data in response]

    def library_subscription_artists(
        self, limit: int = GLOBAL_LIMIT
    ) -> List[YtmusicLibraryArtist]:
        response = self.api.get_library_subscriptions(limit)
        return [YtmusicLibraryArtist(**data) for data in response]

    @ttl_cache(maxsize=CACHE_SIZE, ttl=CACHE_TTL)
    def playlist_info(
        self, playlist_id: str, limit: int = GLOBAL_LIMIT
    ) -> PlaylistInfo:
        return PlaylistInfo(**self.api.get_playlist(playlist_id, limit))

    def liked_songs(self, limit: int = GLOBAL_LIMIT) -> PlaylistInfo:
        return PlaylistInfo(**self.api.get_liked_songs(limit))

    def history(self) -> List[YtmusicHistorySong]:
        response = self.api.get_history()
        return [YtmusicHistorySong(**data) for data in response]

    def create_playlist(
        self,
        title: str,
        description: str,
        privacy_status: YtmusicPrivacyStatus,
        video_ids: List[str] = None,
        source_playlist: str = None,
    ) -> bool:
        response = self.api.create_playlist(
            title, description, privacy_status.value, video_ids, source_playlist
        )
        if not isinstance(response, str):
            return False
        return True

    def add_playlist_items(
        self,
        playlist_id: str,
        video_ids: List[str] = None,
        source_playlist_id: str = None,
    ) -> PlaylistAddItemResponse:
        return PlaylistAddItemResponse(
            **self.api.add_playlist_items(playlist_id, video_ids, source_playlist_id)
        )

    def remove_playlist_items(
        self, playlist_id: str, video_ids: List[dict]
    ) -> Optional[str]:
        # STATUS_SUCCEEDED STATUS_FAILED
        return self.api.remove_playlist_items(playlist_id, video_ids)

    def library_upload_songs(
        self, limit: int = GLOBAL_LIMIT
    ) -> List[YtmusicLibrarySong]:
        response = self.api.get_library_upload_songs(limit)
        return [YtmusicLibrarySong(**data) for data in response]

    def library_upload_artists(
        self, limit: int = GLOBAL_LIMIT
    ) -> List[YtmusicLibraryArtist]:
        response = self.api.get_library_upload_artists(limit)
        return [YtmusicLibraryArtist(**data) for data in response]

    def library_upload_albums(
        self, limit: int = GLOBAL_LIMIT
    ) -> List[YtmusicSearchAlbum]:
        response = self.api.get_library_upload_albums(limit)
        return [YtmusicSearchAlbum(**data) for data in response]

    def upload_song(self, file: str) -> str:
        # STATUS_SUCCEEDED
        return self.api.upload_song_progress(file)

    def delete_upload_song(self, entity_id: str) -> str:
        # STATUS_SUCCEEDED
        return self.api.delete_upload_entity(entity_id)

    def stream_url(
        self, song_info: SongInfo, video_id: str, format_code: int
    ) -> Optional[str]:
        formats = song_info.streamingData.adaptiveFormats
        for f in formats:
            if int(f.itag) == format_code:
                return self._get_stream_url(f, video_id)
        return None

    def check_stream_url(self, url):
        resp = self._session.head(url)
        return resp.status_code != 403


if __name__ == "__main__":
    # noinspection PyUnresolvedReferences

    service = YtmusicService()
    # print(service.upload_song('/home/bruce/Music/阿梓 - 呼吸决定.mp3'))
    print(service.api.delete_upload_entity("t_po_CJT896eTt_a5swEQwrmdzf7_____AQ"))



import logging
from pathlib import Path

from feeluown.app.gui_app import GuiApp
from feeluown.gui.provider_ui import AbstractProviderUi
from feeluown.gui.widgets.login import LoginDialog as LoginDialog_
from feeluown.utils import aio
from feeluown.utils.dispatch import Signal

from fuo_ytmusic.consts import HEADER_FILE
from fuo_ytmusic.headerfile import write_headerfile
from fuo_ytmusic.provider import provider
from fuo_ytmusic.qt_compat import (
    Dialog,
    QFormLayout,
    QInputDialog,
    QLabel,
    QLineEdit,
    QPushButton,
    QVBoxLayout,
    TextSelectableByMouse,
)

logger = logging.getLogger(__name__)


class ProviderUI(AbstractProviderUi):
    def __init__(self, app: GuiApp):
        self._app = app
        self._login_event = Signal()

    @property
    def provider(self):
        return provider

    def get_colorful_svg(self) -> str:
        return (Path(__file__).parent / "assets" / "icon.svg").as_posix()

    def context_menu_add_items(self, menu):
        action = menu.addAction("Switch account")
        action.triggered.connect(lambda: aio.run_afn_ref(self.switch_profile_dialog))

    def login_or_go_home(self):
        if not provider.has_current_user():
            self._dialog = LoginDialog(self.provider)
            self._dialog.login_succeed.connect(self.on_login_succeed)
            self._dialog.show()
        else:
            logger.info("already logged in")
            self.login_event.emit(self, 2)

    @property
    def login_event(self):
        return self._login_event

    def on_login_succeed(self):
        del self._dialog
        self.login_event.emit(self, 1)

    async def switch_profile_dialog(self):
        if not provider.has_current_user():
            self._app.show_msg("Please log in before switching accounts.")
            return
        try:
            profiles = await aio.run_fn(provider.list_profiles)
        except Exception as e:
            self._app.show_msg(f"Failed to fetch account list: {e}")
            return
        if not profiles:
            self._app.show_msg("No available accounts found.")
            return

        items = []
        mapping = {}
        for profile in profiles:
            name = profile.get("accountName") or "Unknown"
            handle = profile.get("channelHandle")
            display = f"{name} ({handle})" if handle else name
            items.append(display)
            mapping[display] = profile

        selected, ok = QInputDialog.getItem(
            self._app,
            "Switch account",
            "Select an account",
            items,
            0,
            False,
        )
        if not ok or not selected:
            return
        profile = mapping.get(selected)
        if not profile:
            return
        try:
            await aio.run_fn(provider.switch_profile, profile.get("accountName"))
        except Exception as e:
            self._app.show_msg(f"Failed to switch account: {e}")
            return
        self._app.show_msg(f"Switched to: {profile.get('accountName')}")
        # Notify UI to refresh current user data/playlists.
        self.login_event.emit(self, 2)


class LoginDialog(LoginDialog_):
    def __init__(self, provider):
        super().__init__()

        self._provider = provider

        self._login_btn = QPushButton("Login")
        self.__hint_label = QLabel(
            "Welcome to the YouTube Music sign-in dialog.\n\n"
            "1) Sign in on music.youtube.com in your browser.\n"
            "2) Open DevTools and trigger a request from Library.\n"
            "3) Copy Authorization and Cookie from that HTTP request.\n"
            "4) Paste them into the fields below, then click Login."
        )
        self.__progress_label = QLabel()
        self.__input_auth_header = QLineEdit()
        self.__input_cookies = QLineEdit()
        self.__hint_label.setTextInteractionFlags(TextSelectableByMouse)
        # Allow users to copy both onboarding hints and runtime status messages.
        self.__progress_label.setTextInteractionFlags(TextSelectableByMouse)
        self.__hint_label.setWordWrap(True)
        self.__progress_label.setWordWrap(True)
        self.setup_ui()

        aio.run_afn_ref(self.try_auto_login)

    def setup_ui(self):
        self.setWindowFlags(self.windowFlags() | Dialog)
        form_layout = QFormLayout()
        form_layout.setContentsMargins(0, 0, 0, 0)
        form_layout.setFieldGrowthPolicy(
            QFormLayout.FieldGrowthPolicy.ExpandingFieldsGrow
        )

        self.__input_auth_header.setPlaceholderText(
            "Example: SAPISIDHASH 1765595295_abcd_u SAPISID1PHASH ..."
        )
        self.__input_cookies.setPlaceholderText("Paste the full cookie value")
        form_layout.addRow("Authorization:", self.__input_auth_header)
        form_layout.addRow("Cookie:", self.__input_cookies)
        self._layout = QVBoxLayout(self)
        self._layout.addWidget(self.__hint_label)
        self._layout.addLayout(form_layout)
        self._layout.addWidget(self.__progress_label)
        self._layout.addWidget(self._login_btn)

        self._login_btn.clicked.connect(lambda: aio.run_afn_ref(self.login))

    def generate_header_file(self, auth: str, cookie: str):
        write_headerfile(auth, cookie, HEADER_FILE)

    async def login(self):
        auth = self.__input_auth_header.text()
        cookie = self.__input_cookies.text()
        if not (auth and cookie):
            self.show_progress("Please input Authorization and Cookie.", color="red")
            return
        self.generate_header_file(auth, cookie)
        self.show_progress("Trying to sign in...", color="green")
        await self.try_login()

    async def try_auto_login(self):
        if HEADER_FILE.exists():
            self.show_progress(
                "Found existing credentials, trying to sign in...",
                color="blue",
            )
            await self.try_login()
        else:
            self.show_progress("Waiting for credentials...", color="blue")

    async def try_login(self):
        try:
            user = await aio.run_fn(self._provider.try_get_user_with_headerfile)
        except Exception as e:
            self.show_progress(f"Sign-in failed: {e}", color="red")
            return

        if user is None:
            self.show_progress(
                "Sign-in failed: credentials may be invalid.",
                color="red",
            )
            return
        self.show_progress(f"Signed in as: {user.name}", color="green")
        self._provider.auth(user)
        self.login_succeed.emit()
        self.hide()

    def show_progress(self, text, color="black"):
        if color is None:
            color = ""
        self.__progress_label.setText(f"<p style='color: {color}'>{text}</p>")


if __name__ == "__main__":
    from feeluown.gui.debug import async_simple_qapp

    with async_simple_qapp():
        dialog = LoginDialog(provider)
        dialog.show()



import json
from http.cookies import SimpleCookie
from pathlib import Path
from typing import Optional

from fuo_ytmusic.consts import HEADER_FILE

# Store selected profile in headerfile using a custom key; this is not part of
# YouTube headers but lets us persist profile selection without extra files.
PROFILE_GAIA_KEY = "X-Feeluown-Profile-Gaia-Id"
# Include both root YouTube domain and Music subdomain explicitly to avoid
# extractor differences when matching cookie scope.
YTDLP_COOKIE_DOMAINS = (".youtube.com", ".music.youtube.com", ".google.com")

# yt-dlp reads cookies from a Netscape cookiefile. We persist this file next to
# ytmusic_header.json and keep it synchronized whenever header auth cookie is
# written/updated, so playback can reuse the same long-lived auth lifecycle.


class YtdlpCookiefileManager:
    """Manage yt-dlp cookiefile generation for a specific headerfile path."""

    def __init__(self, headerfile_path=HEADER_FILE):
        self._headerfile_path = (
            Path(headerfile_path) if headerfile_path is not None else None
        )

    @property
    def headerfile_path(self) -> Optional[Path]:
        return self._headerfile_path

    @property
    def cookiefile_path(self) -> Optional[Path]:
        path = self._headerfile_path
        if path is None or not path.exists():
            return None
        if path.suffix:
            return path.with_suffix(".cookies.txt")
        return Path(f"{path}.cookies.txt")

    def write(self, cookie_value: str) -> Optional[Path]:
        # Write/update the persistent cookiefile used by yt-dlp. Empty cookie
        # means delete the file so stale auth is not reused.
        cookiefile_path = self.cookiefile_path
        if cookiefile_path is None:
            return None

        cookie_jar = SimpleCookie()
        cookie_jar.load(cookie_value or "")
        if not cookie_jar:
            if cookiefile_path.exists():
                cookiefile_path.unlink()
            return None

        lines = [
            "# Netscape HTTP Cookie File",
            "# This file is generated by FeelUOwn ytmusic plugin. Do not edit.",
            "",
        ]
        for morsel in cookie_jar.values():
            for domain in YTDLP_COOKIE_DOMAINS:
                lines.append(
                    f"{domain}\tTRUE\t/\tTRUE\t0\t{morsel.key}\t{morsel.value}"
                )

        cookiefile_path.parent.mkdir(parents=True, exist_ok=True)
        with cookiefile_path.open("w", encoding="utf-8") as handle:
            handle.write("\n".join(lines))
            handle.write("\n")
        return cookiefile_path


def write_headerfile(auth: str, cookie: str, headerfile_path=HEADER_FILE) -> None:
    tpl = {
        "Accept": "*/*",
        "Authorization": auth,
        "Content-Type": "application/json",
        "X-Goog-AuthUser": "0",
        "x-origin": "https://music.youtube.com",
        "Cookie": cookie,
    }
    with headerfile_path.open("w") as handle:
        json.dump(tpl, handle, indent=2, ensure_ascii=True)
    # Keep yt-dlp cookiefile in sync with freshly written headerfile auth.
    YtdlpCookiefileManager(headerfile_path).write(cookie)


def read_headerfile(headerfile_path=HEADER_FILE) -> dict:
    if headerfile_path is None:
        return {}
    try:
        if headerfile_path.exists():
            with open(headerfile_path, "r") as handle:
                data = json.load(handle)
            return data if isinstance(data, dict) else {}
    except Exception:
        return {}
    return {}


def strip_profile_meta(headers: dict) -> dict:
    if not headers:
        return {}
    # Ensure our custom profile marker isn't forwarded as a real request header.
    data = dict(headers)
    data.pop(PROFILE_GAIA_KEY, None)
    return data


def get_profile_gaia_id(headerfile_path=HEADER_FILE) -> Optional[str]:
    # Read persisted profile selection from headerfile (if any).
    data = read_headerfile(headerfile_path)
    return data.get(PROFILE_GAIA_KEY)


def update_profile_gaia_id(gaia_id: Optional[str], headerfile_path=HEADER_FILE) -> bool:
    if headerfile_path is None:
        return False
    # Persist/clear selected profile in headerfile so it survives app restarts.
    data = read_headerfile(headerfile_path)
    original = data.get(PROFILE_GAIA_KEY)
    if gaia_id:
        data[PROFILE_GAIA_KEY] = gaia_id
    else:
        data.pop(PROFILE_GAIA_KEY, None)
    if data.get(PROFILE_GAIA_KEY) == original:
        return False
    with open(headerfile_path, "w") as handle:
        json.dump(data, handle, indent=2, ensure_ascii=True)
    # Profile switch metadata does not change auth cookie, so no cookiefile sync.
    return True


def get_ytdlp_cookiefile_path(headerfile_path=HEADER_FILE) -> Optional[Path]:
    # Backward-compatible helper. New code should use YtdlpCookiefileManager.
    return YtdlpCookiefileManager(headerfile_path).cookiefile_path


def write_ytdlp_cookiefile(
    cookie_value: str, headerfile_path=HEADER_FILE
) -> Optional[Path]:
    # Backward-compatible helper. New code should use YtdlpCookiefileManager.
    return YtdlpCookiefileManager(headerfile_path).write(cookie_value)


def update_headerfile_cookie(cookie_value: str, headerfile_path=HEADER_FILE) -> bool:
    if headerfile_path is None:
        return False
    try:
        if headerfile_path.exists():
            with open(headerfile_path, "r") as handle:
                data = json.load(handle)
        else:
            data = {}
    except Exception:
        data = {}

    changed = data.get("Cookie") != cookie_value
    if changed:
        data["Cookie"] = cookie_value
        with open(headerfile_path, "w") as handle:
            json.dump(data, handle, indent=2, ensure_ascii=True)

    # Keep yt-dlp cookiefile lifecycle aligned with header cookie updates.
    YtdlpCookiefileManager(headerfile_path).write(cookie_value)
    return changed



import json
import logging
from typing import List

from ytmusicapi.constants import YTM_DOMAIN

logger = logging.getLogger(__name__)


class YtmusicProfileManager:
    """Handle profile discovery and switching for authenticated YT Music sessions."""
    def __init__(self, service):
        # Service provides session, auth headers, and API helpers.
        self._service = service
        self._account_override = None
        self._forced_gaia_id = None

    def get_current_account_info(self) -> dict:
        """Return the active profile info, using the account switcher."""
        self._service.api.check_auth()
        if self._account_override is not None:
            return self._account_override

        # Prefer the account switcher endpoint when available.
        switcher = self._get_account_switcher()
        if switcher is not None:
            items = self._find_account_items(switcher)
            datasync_id = self._extract_datasync_id(switcher)
            selected = self._pick_account_item(
                items, datasync_id, self._forced_gaia_id
            )
            if selected is not None:
                info = self._build_account_info(selected)
                if info is not None:
                    if not info.get("channelId"):
                        channel_id = self._extract_channel_id_from_menu()
                        if channel_id:
                            info["channelId"] = channel_id
                    return info

        raise ValueError(
            "No account info available; cookies or authorization may be expired."
        )

    def list_profiles(self) -> List[dict]:
        """List selectable profiles (filtering out items without channel handles)."""
        switcher = self._get_account_switcher()
        if switcher is None:
            return []

        items = self._find_account_items(switcher)
        datasync_id = self._extract_datasync_id(switcher)
        profiles = []
        for item in items:
            name = self._extract_text(item.get("accountName"))
            if not name:
                continue
            channel_handle = self._extract_text(item.get("channelHandle"))
            if not channel_handle:
                continue
            profiles.append(
                {
                    "accountName": name,
                    "channelHandle": channel_handle,
                    "accountPhotoUrl": self._extract_thumbnail_url(
                        item.get("accountPhoto")
                    ),
                    "channelId": self._extract_channel_id(item),
                    "gaiaId": self._extract_obfuscated_gaia_id(item),
                    "isSelected": self._is_item_selected(item, datasync_id),
                }
            )
        return profiles

    def switch_profile(self, account_name: str = None, gaia_id: str = None) -> dict:
        """Switch current profile by name or gaia_id; returns new profile info."""
        if not account_name and not gaia_id:
            self._service.api.set_on_behalf_of_user(None)
            self._account_override = None
            self._forced_gaia_id = None
            return self.get_current_account_info()

        switcher = self._get_account_switcher()
        if switcher is None:
            raise ValueError("No account switcher data available")

        items = self._find_account_items(switcher)
        selected = None
        if gaia_id:
            for item in items:
                if self._extract_obfuscated_gaia_id(item) == gaia_id:
                    selected = item
                    break
        else:
            selected = self._find_account_item_by_name(items, account_name)

        if selected is None:
            raise ValueError("Profile not found")

        selected_gaia_id = self._extract_obfuscated_gaia_id(selected)
        self._service.api.set_on_behalf_of_user(selected_gaia_id)
        self._forced_gaia_id = selected_gaia_id
        self._account_override = self._build_account_info(selected)
        if self._account_override and not self._account_override.get("channelId"):
            channel_id = self._extract_channel_id_from_menu()
            if channel_id:
                self._account_override["channelId"] = channel_id
        return self._account_override

    def _get_account_switcher(self):
        # Account switcher returns the richest profile list, including selection state.
        response = self._request_account_switcher(
            url="https://music.youtube.com/getAccountSwitcherEndpoint",
            origin=YTM_DOMAIN,
        )
        if self._has_account_items(response):
            return response
        return None

    def _request_account_switcher(self, url, origin):
        try:
            response = self._service.api.request_with_auth("GET", url, origin=origin)
            response_text = response.text
        except Exception as e:
            logger.debug("account switcher request failed: %s", e)
            return None

        if response.status_code >= 400:
            logger.debug(
                "account switcher request failed with status %s",
                response.status_code,
            )
            return None

        # Some responses update auth cookies; keep them in sync.
        self._service.api.update_auth_cookie_from_response(response)

        payload = self._strip_xssi_prefix(response_text)
        try:
            return json.loads(payload)
        except Exception as e:
            logger.debug("account switcher parse failed: %s", e)
            return None

    @classmethod
    def _strip_xssi_prefix(cls, text):
        if not isinstance(text, str):
            return text
        if text.startswith(")]}'"):
            newline = text.find("\n")
            if newline != -1:
                return text[newline + 1 :]
            return text[4:]
        return text

    def _extract_channel_id_from_menu(self):
        # account_menu does not always include it, but can expose a channel browseId.
        try:
            response = self._service.api.send_api_request("account/account_menu", {})
        except Exception as e:
            logger.debug("account_menu request failed: %s", e)
            return None
        return self._find_channel_browse_id(response)

    @classmethod
    def _find_account_items(cls, data):
        items = []
        if isinstance(data, dict):
            for key in ("accountItem", "accountItemRenderer"):
                item = data.get(key)
                if isinstance(item, dict):
                    items.append(item)
            for value in data.values():
                items.extend(cls._find_account_items(value))
        elif isinstance(data, list):
            for value in data:
                items.extend(cls._find_account_items(value))
        return items

    @classmethod
    def _find_channel_browse_id(cls, data):
        # Look for a user channel browseId in any nested browseEndpoint.
        if isinstance(data, dict):
            browse_endpoint = data.get("browseEndpoint")
            if isinstance(browse_endpoint, dict):
                browse_id = browse_endpoint.get("browseId")
                if browse_id and str(browse_id).startswith("UC"):
                    return browse_id
                page_type = (
                    browse_endpoint.get("browseEndpointContextSupportedConfigs", {})
                    .get("browseEndpointContextMusicConfig", {})
                    .get("pageType")
                )
                if page_type == "MUSIC_PAGE_TYPE_USER_CHANNEL" and browse_id:
                    return browse_id
            for value in data.values():
                found = cls._find_channel_browse_id(value)
                if found:
                    return found
        elif isinstance(data, list):
            for value in data:
                found = cls._find_channel_browse_id(value)
                if found:
                    return found
        return None

    @classmethod
    def _pick_account_item(cls, items, datasync_id=None, gaia_id=None):
        # Explicit gaia_id wins; otherwise rely on selection flags or datasync_id.
        if gaia_id:
            for item in items:
                if cls._extract_obfuscated_gaia_id(item) == gaia_id:
                    return item
        for item in items:
            if item.get("isSelected") or item.get("isCurrent") or item.get("isActive"):
                return item
        if datasync_id:
            for item in items:
                if cls._extract_obfuscated_gaia_id(item) == datasync_id:
                    return item
        return None

    @classmethod
    def _find_account_item_by_name(cls, items, name):
        if not name:
            return None
        for item in items:
            if cls._extract_text(item.get("accountName")) == name:
                return item
        return None

    @classmethod
    def _extract_datasync_id(cls, response):
        if not isinstance(response, dict):
            return None
        response_context = response.get("responseContext", {})
        main_context = response_context.get("mainAppWebResponseContext", {})
        datasync_id = main_context.get("datasyncId")
        if not datasync_id or not isinstance(datasync_id, str):
            return None
        return datasync_id.split("||", 1)[0]

    @classmethod
    def _extract_obfuscated_gaia_id(cls, item):
        if not isinstance(item, dict):
            return None
        service_endpoint = item.get("serviceEndpoint", {})
        select_endpoint = service_endpoint.get("selectActiveIdentityEndpoint", {})
        tokens = select_endpoint.get("supportedTokens", [])
        if not isinstance(tokens, list):
            return None
        for token in tokens:
            account_state = token.get("accountStateToken")
            if isinstance(account_state, dict):
                obfuscated_id = account_state.get("obfuscatedGaiaId")
                if obfuscated_id:
                    return obfuscated_id
        return None

    @classmethod
    def _has_account_items(cls, response):
        if response is None:
            return False
        return bool(cls._find_account_items(response))

    @classmethod
    def _extract_text(cls, value):
        if isinstance(value, dict):
            runs = value.get("runs")
            if isinstance(runs, list) and runs:
                text = "".join(
                    run.get("text", "") for run in runs if isinstance(run, dict)
                ).strip()
                if text:
                    return text
            simple_text = value.get("simpleText")
            if simple_text:
                return simple_text
        if isinstance(value, str):
            return value
        return None

    @classmethod
    def _build_account_info(cls, item):
        account_name = cls._extract_text(item.get("accountName"))
        if not account_name:
            return None
        channel_handle = cls._extract_text(item.get("channelHandle"))
        channel_id = cls._extract_channel_id(item)
        gaia_id = cls._extract_obfuscated_gaia_id(item)
        account_photo_url = cls._extract_thumbnail_url(
            item.get("accountPhoto")
        )
        return {
            "accountName": account_name,
            "channelHandle": channel_handle,
            "accountPhotoUrl": account_photo_url,
            "channelId": channel_id,
            "gaiaId": gaia_id,
        }

    @classmethod
    def _extract_thumbnail_url(cls, account_photo):
        if not isinstance(account_photo, dict):
            return None
        thumbnails = account_photo.get("thumbnails")
        if isinstance(thumbnails, list) and thumbnails:
            return thumbnails[0].get("url")
        return None

    @classmethod
    def _extract_channel_id(cls, item):
        if not isinstance(item, dict):
            return None
        return item.get("channelId")

    @classmethod
    def _is_item_selected(cls, item, datasync_id=None):
        if item.get("isSelected") or item.get("isCurrent") or item.get("isActive"):
            return True
        if datasync_id:
            return cls._extract_obfuscated_gaia_id(item) == datasync_id
        return False



import logging
from typing import List, Optional, Sequence, Tuple

from feeluown.library import (
    BriefAlbumModel,
    BriefSongModel,
    BriefVideoModel,
    Collection,
    CollectionType,
    PlaylistModel,
)

from fuo_ytmusic.models import YtmusicHomePlaylist, YtmusicHomeSong, YtmusicSearchAlbum

logger = logging.getLogger(__name__)


class YtmusicHomeRecommendationBuilder:
    DEFAULT_SECTION_TITLE = "Recommendations"
    TYPE_TITLE_SUFFIXES = {
        CollectionType.only_songs: "Songs",
        CollectionType.only_playlists: "Playlists",
        CollectionType.only_albums: "Albums",
        CollectionType.only_videos: "Videos",
    }

    def __init__(self, source: str):
        self._source = source

    @staticmethod
    def normalize_limit(limit: Optional[int], default_limit: int) -> int:
        if limit is None:
            return default_limit
        try:
            normalized = int(limit)
        except (TypeError, ValueError):
            return default_limit
        return normalized if normalized > 0 else default_limit

    def build_collections(self, sections: Sequence[dict]) -> List[Collection]:
        collections: List[Collection] = []
        for section in sections or []:
            if not isinstance(section, dict):
                continue
            contents = section.get("contents")
            if not isinstance(contents, list):
                continue

            section_title = section.get("title") or self.DEFAULT_SECTION_TITLE
            typed_collections = self._build_typed_collections(contents)
            if not typed_collections:
                continue
            if len(typed_collections) == 1:
                section_type, models = typed_collections[0]
                collections.append(
                    Collection(name=section_title, type_=section_type, models=models)
                )
                continue
            for section_type, models in typed_collections:
                suffix = self.TYPE_TITLE_SUFFIXES[section_type]
                collections.append(
                    Collection(
                        name=f"{section_title} · {suffix}",
                        type_=section_type,
                        models=models,
                    )
                )
        return collections

    def _build_typed_collections(
        self, contents: Sequence[dict]
    ) -> List[Tuple[CollectionType, list]]:
        type_builder_pairs = [
            (CollectionType.only_songs, self._build_home_songs),
            (CollectionType.only_playlists, self._build_home_playlists),
            (CollectionType.only_albums, self._build_home_albums),
            (CollectionType.only_videos, self._build_home_videos),
        ]
        collections = []
        for section_type, builder in type_builder_pairs:
            models = builder(contents)
            if models:
                collections.append((section_type, models))
        return collections

    @staticmethod
    def _iter_valid_contents(contents: Sequence[dict]):
        for content in contents or []:
            if isinstance(content, dict):
                yield content

    @staticmethod
    def _detect_home_item_type(content: dict):
        playlist_id = content.get("playlistId")
        video_id = content.get("videoId")
        artists = content.get("artists")
        album = content.get("album")

        if playlist_id and not video_id:
            # Includes watch playlist and mix/radio-like playlist rows.
            return CollectionType.only_playlists
        if video_id:
            # Some cards carry both videoId and playlistId. If there is no song
            # metadata, treat them as playlist-like mix cards.
            if playlist_id and not artists and not album:
                return CollectionType.only_playlists
            # YTMusic home videos usually include a view-count field.
            if content.get("views"):
                return CollectionType.only_videos
            return CollectionType.only_songs
        if playlist_id:
            return CollectionType.only_playlists
        if content.get("browseId") and content.get("artists") is not None:
            # Artists also carry browseId; require artists field to identify albums.
            return CollectionType.only_albums
        return None

    def _build_home_songs(self, contents: Sequence[dict]) -> List[BriefSongModel]:
        songs: List[BriefSongModel] = []
        seen_video_ids = set()
        for content in self._iter_valid_contents(contents):
            if self._detect_home_item_type(content) != CollectionType.only_songs:
                continue
            video_id = content.get("videoId")
            if not video_id or video_id in seen_video_ids:
                continue
            try:
                song = YtmusicHomeSong(**content).v2_brief_model()
            except Exception as e:
                logger.warning("skip invalid home song item(%s): %s", video_id, e)
                continue
            if not song.identifier:
                continue
            seen_video_ids.add(song.identifier)
            songs.append(song)
        return songs

    def _build_home_playlists(self, contents: Sequence[dict]) -> List[PlaylistModel]:
        playlists: List[PlaylistModel] = []
        seen_playlist_ids = set()
        for content in self._iter_valid_contents(contents):
            if self._detect_home_item_type(content) != CollectionType.only_playlists:
                continue
            playlist_id = content.get("playlistId")
            if not playlist_id or playlist_id in seen_playlist_ids:
                continue
            try:
                playlist = YtmusicHomePlaylist(**content).v2_model()
            except Exception as e:
                logger.warning("skip invalid home playlist item(%s): %s", playlist_id, e)
                continue
            if not playlist.identifier:
                continue
            seen_playlist_ids.add(playlist.identifier)
            playlists.append(playlist)
        return playlists

    def _build_home_albums(self, contents: Sequence[dict]) -> List[BriefAlbumModel]:
        albums: List[BriefAlbumModel] = []
        seen_browse_ids = set()
        for content in self._iter_valid_contents(contents):
            if self._detect_home_item_type(content) != CollectionType.only_albums:
                continue
            browse_id = content.get("browseId")
            if not browse_id or browse_id in seen_browse_ids:
                continue
            try:
                album = YtmusicSearchAlbum(**content).v2_brief_model()
            except Exception as e:
                logger.warning("skip invalid home album item(%s): %s", browse_id, e)
                continue
            if not album.identifier:
                continue
            seen_browse_ids.add(album.identifier)
            albums.append(album)
        return albums

    def _build_home_videos(self, contents: Sequence[dict]) -> List[BriefVideoModel]:
        videos: List[BriefVideoModel] = []
        seen_video_ids = set()
        for content in self._iter_valid_contents(contents):
            if self._detect_home_item_type(content) != CollectionType.only_videos:
                continue
            video_id = content.get("videoId")
            if not video_id or video_id in seen_video_ids:
                continue
            artists = content.get("artists")
            artists_name = " / ".join(
                artist.get("name", "")
                for artist in (artists or [])
                if isinstance(artist, dict) and artist.get("name")
            )
            videos.append(
                BriefVideoModel(
                    identifier=video_id,
                    source=self._source,
                    title=content.get("title") or "",
                    artists_name=artists_name,
                    duration_ms=content.get("duration") or "",
                )
            )
            seen_video_ids.add(video_id)
        return videos



from typing import Optional

WATCH_NEXT_RENDERER_PATH = (
    "contents",
    "singleColumnMusicWatchNextResultsRenderer",
    "tabbedRenderer",
    "watchNextTabbedResultsRenderer",
)


def build_watch_playlist_body(video_id: str) -> dict:
    return {
        "enablePersistentPlaylistPanel": True,
        "isAudioOnly": True,
        "tunerSettingValue": "AUTOMIX_SETTING_NORMAL",
        "videoId": video_id,
        "playlistId": f"RDAMVM{video_id}",
        "watchEndpointMusicSupportedConfigs": {
            "watchEndpointMusicConfig": {
                "hasPersistentPlaylistPanel": True,
                "musicVideoType": "MUSIC_VIDEO_TYPE_ATV",
            }
        },
    }


def extract_lyrics_browse_id(watch_response) -> Optional[str]:
    tabs = watch_response
    for key in WATCH_NEXT_RENDERER_PATH:
        tabs = tabs[key]
    tabs = tabs["tabs"]

    try:
        browse_id = tabs[1]["tabRenderer"]["endpoint"]["browseEndpoint"]["browseId"]
    except (IndexError, KeyError, TypeError):
        return None
    if not isinstance(browse_id, str):
        raise TypeError(f"unexpected lyrics browse id type: {type(browse_id)}")
    return browse_id


def timestamped_lyrics_to_lrc(lyrics_payload) -> Optional[str]:
    lyrics = lyrics_payload["lyrics"]
    if not isinstance(lyrics, list):
        return None

    lines = [format_lyric_line(line) for line in lyrics]
    if not lines:
        return None
    return "\n".join(lines)


def format_lyric_line(line) -> str:
    return f"{format_lrc_timestamp(line.start_time)}{line.text}"


def format_lrc_timestamp(milliseconds: int) -> str:
    if not isinstance(milliseconds, int):
        raise TypeError(f"unexpected lyric start_time type: {type(milliseconds)}")
    total_ms = max(0, milliseconds)
    minutes = total_ms // 60000
    seconds = (total_ms % 60000) // 1000
    centiseconds = (total_ms % 1000) // 10
    return f"[{minutes:02d}:{seconds:02d}.{centiseconds:02d}]"



from pathlib import Path
from feeluown.consts import DATA_DIR
HEADER_FILE = Path(DATA_DIR) / "ytmusic_header.json"
REQUIRED_COOKIE_FIELDS = [
    "HSID",
    "SSID",
    "APISID",
    "SAPISID",
    "__Secure-3PAPISID",
    "LOGIN_INFO",
    "__Secure-1PAPISID",
    "SID",
    "__Secure-1PSID",
    "__Secure-3PSID",
    "__Secure-3PSIDCC",
]
